extends KinematicBody

export(NodePath) onready var body = get_node(body) as Spatial
export(NodePath) onready var head = get_node(head) as Spatial
export(NodePath) onready var cam = get_node(cam) as Camera
onready var col_stand = $CollisionStand
onready var col_crouch = $CollisionCrouch
export(NodePath) onready var tween = get_node(tween) as Tween

var mouse_sensitivity := 20.0

var stop_speed := 4.0
var move_speed := 10.0
var max_slope_angle := 45
var min_step_angle := 70 # must be less than 90, must be more than <play_with_to_find_such_number>

var gravity := 60.0
var max_fall_speed := 50.0
var jump_height := 2.0

var accel_ground := 10.0
var accel_air := 3.0
var accel_water := 4.0

var friction_ground := 6.0
var friction_water := 1.0

var step_height := 1.0

var queue_jump := false
var snap_vector := Vector3.ZERO
var grounded := false
var wall := false
var crouch := false
var on_ladder := false
var swimming := false

const cr_sp = 0.3
const ldr_sp = 0.5

onready var head_offset := [head.transform.origin.y,head.transform.origin.z]

var velocity := Vector3.ZERO

onready var current_shape = col_stand

func _ready():
	Input.set_mouse_mode(Input.MOUSE_MODE_CAPTURED)

func get_move_direction() -> Vector3:
	
	var input_dir := Vector2(
		Input.get_axis("move_forward","move_backward"),
		Input.get_axis("move_left", "move_right")
	).clamped(1.0)
	
	if !on_ladder and !swimming:
		return input_dir.x * body.global_transform.basis.z + input_dir.y * body.global_transform.basis.x
	else:
		return input_dir.x * head.global_transform.basis.z + input_dir.y * body.global_transform.basis.x

func _input(event):
	if event is InputEventMouseMotion:
		rotate_look(event.relative * 0.001 * mouse_sensitivity)
	
	if Input.is_action_just_pressed("crouch"):
		if grounded:
			if !crouch:
				col_stand.disabled = true
				col_crouch.disabled = false
				crouch_tween()
				crouch = true
				
			else:
				if overlap_check() == []:
					col_stand.disabled = false
					col_crouch.disabled = true
					crouch_tween()
					crouch = false
		# assign current hull
		current_shape = col_stand if not col_stand.disabled else col_crouch
	
	if Input.is_action_just_pressed("move_jump") and !crouch:
		queue_jump = true
	if Input.is_action_just_released("move_jump") and !crouch:
		queue_jump = false
	
	if Input.is_action_just_pressed("exit"):
		get_tree().quit()

func crouch_tween():
	tween.stop_all()
	tween.interpolate_property(head,"translation",
	head.translation,
	Vector3(0,head_offset[0],head_offset[1]) if crouch else Vector3(0,-0.75,head_offset[1]),
	0.2,
	Tween.EASE_OUT,Tween.EASE_OUT)
	tween.start()

func rotate_look(amount : Vector2) -> void:
	body.rotation.y -= amount.x
	head.rotation.x = clamp(head.rotation.x - amount.y, -PI * 0.5, PI * 0.5)

func _physics_process(delta):
	grounded = is_on_floor()
	wall = is_on_wall()
	
	velocity.y = max(-max_fall_speed, velocity.y - (gravity if (!grounded and !on_ladder) else 0.0) * delta)
	
	var h_target_dir : Vector3 = get_move_direction()
	if !on_ladder:
		if grounded:
			apply_friction(delta)
			accelerate(delta, h_target_dir, move_speed * (cr_sp if crouch else 1.0), accel_ground)
			
			snap_vector = -get_floor_normal() - get_floor_velocity() * delta
			
			# Workaround for sliding down after jump on slope
			if velocity.y < 0:
				velocity.y = 0
				
			if queue_jump:
				velocity.y = sqrt(2 * jump_height * abs(gravity))
				queue_jump = false
				snap_vector = Vector3.ZERO
			
			else: #kiwi: not gonna step while jumping (and other hard_surface stuff)
				up_stepping(delta)
		else:
			# Workaround for 'vertical bump' when going off platform
			if snap_vector != Vector3.ZERO && velocity.y != 0:
				velocity.y = 0
				
			accelerate(delta, h_target_dir, move_speed, accel_air)
			snap_vector = Vector3.ZERO
	elif on_ladder:
		snap_vector = Vector3.ZERO
		apply_friction(delta)
		accelerate(delta, h_target_dir, move_speed * ldr_sp, accel_ground)
		if test_move(global_transform,Vector3(velocity.x,0,velocity.z),true):
			velocity.x  = 0

	velocity = move_and_slide_with_snap(velocity, snap_vector, Vector3.UP, true, 4, deg2rad(max_slope_angle), false)

func accelerate(delta : float, p_target_dir : Vector3, p_target_speed : float, p_accel : float):
	var current_speed : float = velocity.dot(p_target_dir)
	var add_speed : float = p_target_speed - current_speed
	#old code
	if add_speed > 0:
		var accel_speed : float = min(add_speed, p_accel * delta * p_target_speed)
		velocity += p_target_dir * accel_speed
#		print("starting_speed: " + str(move_speed))
#		print("output_speed: " + str(velocity.length()))

func apply_friction(delta : float):
	
	var speed : float = velocity.length()
	if is_zero_approx(speed):
		velocity = Vector3.ZERO
		return
	
	var drop : float = 0.0
	var control : float = max(speed, stop_speed)
	
	# ground friction
	if grounded or on_ladder:
		drop += control * friction_ground * delta
	
	# water friction not implemented
	var new_speed : float = max(0.0, speed - drop)
	new_speed /= speed
	
	velocity *= new_speed

func overlap_check():
	var space: PhysicsDirectSpaceState = get_world().direct_space_state as PhysicsDirectSpaceState
	
	var params: = PhysicsShapeQueryParameters.new()
	params.set_shape(col_stand.shape)
	params.set_transform(global_transform)
	params.set_exclude([self])
	
	var castResult = space.intersect_shape(params)
	return castResult

func up_stepping( delta:float ):
	var space_state : PhysicsDirectSpaceState = get_world().direct_space_state as PhysicsDirectSpaceState
	var param := PhysicsShapeQueryParameters.new()
	var motion :Vector3 = velocity * delta
	var reminder :Vector3
#	var extra_margin := 0.1
	var extra_margin := 0.0
	var current_shape = col_stand if !col_stand.disabled else col_crouch
	
	param.collision_mask = self.collision_mask
	param.set_shape( current_shape.shape )
	param.transform = current_shape.global_transform
	param.margin = current_shape.shape.margin
	param.exclude = [ self ]

	var trace := space_state.cast_motion( param, motion )

	if not (trace[0] == 1.0 and trace[1] == 1.0): # frontal collision
		
		var front_normal = space_state.get_rest_info(param).get("normal").y if space_state.get_rest_info(param).get("normal") != null else 0
#		print(front_normal)
		param.transform.origin += motion * trace[1]     # I think if reminder become too small, trace step down will miss the collision later. With this, change shape will collide with step but cast_motion ignore it based on document, no worries
		reminder = motion - (motion * trace[1])
		motion = Vector3.UP * (step_height + extra_margin)
		trace = space_state.cast_motion( param, motion )

		if trace[0] == 1.0 and trace[1] == 1.0: # no ceiling collision
			param.transform.origin += motion
			trace = space_state.cast_motion( param, reminder )

			if trace[0] > 0.0: # if there's space forward..
				param.transform.origin += reminder * trace[1]

				# that code_block could be replaced by a recursive function
				if not (trace[0] == 1.0 and trace[1] == 1.0):     # frontal collision with next-step → cast_up
					motion = Vector3.UP * (step_height + extra_margin)
					trace = space_state.cast_motion( param, motion )

					if trace[0] == 1.0 and trace[1] == 1.0: # no ceiling collision → cast_forward
						param.transform.origin += motion
						reminder -= reminder * trace[1]
						trace = space_state.cast_motion( param, reminder )

						if trace[0] == 1.0 and trace[1] == 1.0:
							param.transform.origin += reminder

				motion = Vector3.DOWN * (step_height + extra_margin)
				trace = space_state.cast_motion( param, motion )

				if not (trace[0] == 1.0 and trace[1] == 1.0): # collision with step-surface
					param.transform.origin += motion * trace[1]
					var collision     := space_state.get_rest_info(param)

#					if not collision.empty() and not Vector3.UP.angle_to( collision.get("normal") ) > max_slope_angle:      # avoid climb "mountains"
#					if not collision.empty() and collision.get("normal").y > 0.99:      # avoid climb "mountains"
					if not collision.empty():
#						if collision.get("normal").y < 0.99 and get_floor_normal().y < 0.99:
#						if collision.get("normal").y < 0.99 and (stepify(collision.get("normal").y, 0.1) == stepify(get_floor_normal().y, 0.1)):
#						if collision.get("normal").y < 0.99 and stepify(front_normal, 0.01) == stepify(get_floor_normal().y, 0.01):
						if collision.get("normal").y < 0.99 and stepify(front_normal, 0.01) == stepify(get_floor_normal().y, 0.01) and stepify(collision.get("normal").y, 0.01) == stepify(get_floor_normal().y, 0.01):
							print(front_normal, "+ ", get_floor_normal().y, "+ ", collision.get("normal").y)
#							print(stepify(collision.get("normal").y, 0.1) , "+", stepify(get_floor_normal().y, 0.1))
							print("ramp!")
							return
#						if not collision.get("point").y - self.global_transform.origin.y > step_height:
#						if not param.transform.origin - self.transform.origin.y > step_height
#						if not param.transform.origin - self.global_transform.origin.y > step_height
						if not self.global_transform.origin.distance_to(param.transform.origin) > step_height:
#						if not self.transform.origin.distance_to(param.transform.origin) > step_height:
#							print(self.global_transform.origin.distance_to(param.transform.origin))
#							print(self.global_transform.origin.y - param.transform.origin.y)
							self.transform.origin = param.transform.origin
							self.transform.origin.y -= current_shape.transform.origin.y
							print("upstep!")
							# using Zylann's plugin to visualize shape-casting
							DebugDraw.draw_box( param.transform.origin, Vector3( current_shape.shape.radius, current_shape.shape.height, current_shape.shape.radius ) * 2, Color(0,1,0) )

